cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c  SVM-QP, written by Katya Scheinberg, IBM Research, 2005.
c  to contact the author email at katyas'at'us'dot'ibm'dot'com
c
c Copyright (C) 2005 International Business Machines.
c All Rights Reserved.
c This code is published under the Common Public License.
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine inc_actvset(points, ipoints, lpnt, lipnts, ipntr, a,  
     .                       c, x, y, k, nrnk, shftk, dgrk, maxns,   
     .                       maxn, ninit, usesift, maxsift0, maxsiftc,  
     .                       fullmem, tol, maxit, cornrsln, iprint, 
     .                       wrk, lwrk, iwrk, liwrk)
c
c  This routine initialized the active set method and calls it in the incremental mode,
c  if it is applicable, adding one point at a time.
c    
c
c  points
c  ipoints  - arrays that store the information about the data points
c  ipntr 
c  a        - labels
c  c        - penatly parameter
c  x        - vector of lagrange multipliers
c  y        - negative bias
c  nrnk     - maximum number of linearly indep points
c  shftk    - kernel parameter (shift of sigma)
c  dgrk     - kernel parameter(degree, or indication that RBF is used)
c  wrk, iwrk- working arrays
c 
      double precision  a(maxn), c, x(maxn), y, points(lpnt), 
     .                  wrk(lwrk), shftk

      integer          iwrk(liwrk), liwrk, n,  k, i, r, nrnk, dgrk,
     .                 maxns, lwrk, maxn, ninit, fullmem,
     .                 ipoints(lipnts), ipntr(maxn), lpnt, nbatch, 
     .                 iprint, maxsift0, maxsiftc, lipnts
      logical          usesift, cornrsln, debug, initlz, batch, testerr
c
c  memory pointers
c
      integer          ia,  ix, is, ixi, 
     *                 icurw,  kk, m, j, iperm, 
     *                 ig, itrc, iactv, iatzero, iatone, icuriw, kk1, 
     *                 kk2, kk3, jj, l1, l2, ii0, iic, iis, nc, maxit,
     *                 ns, n0, iqy0, iqxs, ixs, iqys, iqx0, iqyc, iqxc,
     *                 iter, iter1, iter2, ncapx, n0apx, nsapx,
     *                 itertotal, l, lis, lic, li0, 
     *                 ii0ind, iicind, li0ind, licind, ii0disc,
     *                 li0disc, n0disc, ncdisc, iicdisc, licdisc, iiqs,
     *                 liqs, lqs, iqs,  iirm, lirm, ir, lr,
     *                 lidisc, iidisc, nremain,  thresh, nns, nn0, nnc,
     *                 erro, nn, iqc, lqc, inminslk, itest,
     *                 ibeg, iminslk, iminsur, iiminslk, iiminsur, 
     *                 nposmis, nnegmis, nposcor, nnegcor
c
c  other local variables
c
      double precision  dy, tmp, theta, tol, boxtol, shift, dir, slk,
     .                  totalt,  val,
     .                  currt, big,  ac, minslk
      double precision  gkernlsp
      external          gkernlsp
c
c  common block for the kernel function
c
      logical          sparse
      integer          ncalls
      double precision timek
      common /sprs/    sparse, ncalls, timek
Ctime      currt=mclock()
c
c  memory pointers
c
      ixs     = 1
      ir      = ixs + maxn
      iqxs    = ir + maxn
      iqys    = iqxs + maxn
      iqxc    = iqys + maxn
      iqyc    = iqxc + maxn
      iqx0    = iqyc + maxn
      iqy0    = iqx0 + maxn
      iqs     = iqy0 + maxn
      iqc     = iqs  + maxn*maxns
      ig      = iqc  + maxn
      iminslk = ig   + maxns*maxns
      iminsur = iminslk + maxsift0
      icurw   = iminsur + maxsiftc
      
      lqs     = iqs   - 1
      lqc     = iqc   - 1

      iis     = 1
      iic     = iis + maxn
      ii0     = iic + maxn
      ii0ind  = ii0 +maxn
      iicind  = ii0ind + maxn
      iidisc  = iicind + maxn
      ii0disc = iidisc + maxn
      iicdisc = ii0disc + maxn
      iiqs    = iicdisc + maxn
      iirm    = iiqs + maxns
      iperm   = iirm + maxn
      iiminslk = iperm + maxns
      iiminsur = iiminslk + maxsift0
      itest    = iiminsur + maxsiftc
      icuriw   = itest + maxn


      lis     = iis - 1
      lic     = iic - 1
      li0     = ii0 - 1
      licind  = iicind - 1
      li0ind  = ii0ind - 1
      lidisc  = iidisc - 1
      li0disc = ii0disc - 1
      licdisc = iicdisc - 1
      liqs    = iiqs - 1
      lirm    = iirm - 1
      lr      = ir - 1
    
c
c  check if memory is sufficient, if not, print a message and exit
c
      if (icurw  .gt. lwrk ) then
        if (iprint.ge.0) write(6, 1000) icurw 
        goto 999
      endif

      if (icuriw + maxn  .gt. liwrk ) then
        if (iprint.ge.0) write(6, 1100) icurw + maxns
        goto 999
      endif

c
c  Initial setting
c
      testerr=.false.
      batch=.false.
      nbatch=1
      big = 1.0d20
      maxit = 1000000
      if ( testerr) then
        n=maxn-maxn/10
      else
        n=ninit
      endif
      nc=0
      n0=0
      ns=0
c
c  find the initial partition of variables  into sets Is, I0, Ic.
c
        do 10 i=1, n
        if ( x(i) .lt. 100*tol) then
          x(i)=0.0d0
          iwrk(itest+i)=0
          iwrk(ii0+n0)=i
          iwrk(ii0ind+n0)=0
          n0=n0+1
        else if ( x(i) .gt. c-100*tol) then 
          x(i)=c
          iwrk(itest+i)=2
          iwrk(iic+nc)=i
          iwrk(iicind+nc)=0
          nc=nc+1
        else
          iwrk(itest+i)=1
          iwrk(iis+ns)=i
          wrk(ixs+ns)=x(i)
          ns=ns+1
        endif
 10   continue
      n0apx=n0
      ncapx=nc
      nsapx=ns
c
c  If desired project the current approximate solution on the bounds, so that the 
c  active set will start from a corner solution.
c
      if (cornrsln) then
        ac=0.0
        do 15 i=1,nc
          ac=ac+a(iwrk(iic+i-1))
 15     continue
        if (ac .gt. 0 ) then 
          j=1
        else
          j=-1
        endif 
          i=0 
          do while (j*ac .ge. 1)
            if ( j*a(iwrk(iis+i)).le. -1 ) then
              iwrk(iic+nc)=iwrk(iis+i)
              iwrk(iicind+nc)=0
              x(iwrk(iic+nc))=c
              ac=ac+a(iwrk(iic+nc))
              nc=nc+1
            else
              iwrk(ii0+n0)=iwrk(iis+i)
              iwrk(ii0ind+n0)=0
              x(iwrk(ii0+n0))=0.0d0
              n0=n0+1
            endif
            i=i+1
         end do 
         do 16 kk=i,ns-1
          iwrk(ii0+n0)=iwrk(iis+kk)
          iwrk(ii0ind+n0)=0
          x(iwrk(ii0+n0))=0.0d0
          n0=n0+1
 16      continue
         ns=0
       endif
c
c  initialize array of indices indication which points out of all points that should be
c  added are remaining in the active pool 
c
       do 21 i=1, maxn-n
         iwrk(lirm+i)=i+n
 21     continue 
c
c  initialize array of indices pointing where a columns of Qs is stored
c 
       do 22 i=1, maxns
         iwrk(liqs+i)=i
 22     continue 
c
c compute current vector Qc*c
c
       do 24 i = 1, ns
        l=lqc+iwrk(lis+i)
        wrk(l)=0.0d0
        do 23 j = 1, nc
          wrk(l)=wrk(l)+gkernlsp(points, ipoints, lpnt, ipntr, a,  
     .                l-lqc, iwrk(lic+j),  maxn, k, shftk, dgrk )
 23    continue
       wrk(l)= wrk(l)*c
 24   continue  
       do 26 i = 1, nc
        l=lqc+iwrk(lic+i)
        wrk(l)=0.0d0
        do 25 j = 1, nc
          wrk(l)=wrk(l)+gkernlsp(points, ipoints, lpnt, ipntr, a,  
     .                l-lqc, iwrk(lic+j),  maxn, k, shftk, dgrk )
 25     continue
       wrk(l)= wrk(l)*c
 26    continue  
      do 28 i = 1, n0
        l=lqc+iwrk(li0+i)
        wrk(l)=0.0d0
        do 27 j = 1, nc
          wrk(l)=wrk(l)+gkernlsp(points, ipoints, lpnt, ipntr, a, 
     .                l-lqc, iwrk(lic+j), maxn, k,shftk, dgrk )
 27     continue
        wrk(l)= wrk(l)*c
 28   continue
c
c  Compute intial Qs
c
      do 35 i=1, ns   
        if ( fullmem .eq. 2 ) then     
          call addcolqs(points,  ipoints, lpnt, ipntr, a, 
     .                    wrk(iqs), iwrk(liqs+i), iwrk(iis),  i, n,
     .                    maxn, maxns,  k, shftk, dgrk, wrk, .false.)
        else 
          call addcolqsmin(points,  ipoints, lpnt, ipntr, a,  
     .                    wrk(iqs), iwrk(liqs+i), 
     .                    iwrk(iis),  iwrk(iic), iwrk(ii0), i, nc, 
     .                    n0,  maxn, maxns, k, shftk, dgrk)   
        endif      
 35   continue 
 
c
c  initialize some values
c
      itertotal=0
      n0disc=0
      ncdisc=0
      thresh=100
      ibeg=1
      nrnk=0
      do 40 i=1,ns
        iwrk(iperm+i-1)=i
 40   continue  
      initlz=.false.
ccccccccccccccccccccccccccccccc
c  START THE MAIN ITERATION
ccccccccccccccccccccccccccccccc
      do 500 iter1=1, maxn   


 100    call actvset(points,  ipoints, lpnt, ipntr, a, c, x, y, 
     *         iwrk(iis), iwrk(iic), iwrk(ii0), ns, nc, n0, n,
     *         k, ibeg, nrnk, maxit, wrk(ixs), wrk(iqxs), wrk(iqys),
     *         wrk(iqxc), wrk(iqyc), wrk(iqx0), wrk(iqy0),  
     *         iwrk(iiqs), wrk(iqs), wrk(iqc),  thresh,
     *         iwrk(ii0ind), iwrk(iicind), iwrk(iicdisc),
     *         ncdisc, iwrk(ii0disc), n0disc, shftk, dgrk,  
     *         maxns, maxn, wrk(ig), iwrk(iperm),  fullmem, 
     *         usesift,  wrk(iminslk), iwrk(iiminslk), maxsift0, 
     *         wrk(iminsur), iwrk(iiminsur), maxsiftc, 
     *         initlz, tol, iter, itertotal, iprint, wrk(icurw), 
     *         lwrk-icurw+1, iwrk(icuriw), liwrk-icuriw+1)


cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
ccccccc
ccccccc       check for correctness  of the returned solution
ccccccc
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      debug=.false.
      if (debug .and. iprint. ge. 0) then
      do 719 i= 1, maxn
        wrk(lr+i)= -1.0d0-a(i)*y
 719   continue  
      do 720 i = 1, ns
         slk=wrk(lr+iwrk(lis+i))
         do 721 j=1,ns
           slk= slk+wrk(ixs+j-1)*gkernlsp(points, ipoints, lpnt,    
     .          ipntr, a, iwrk(lis+i), iwrk(lis+j), maxn, k, shftk, 
     .          dgrk )
 721      continue  
         do 722 j=1,nc
           slk= slk+c*gkernlsp(points,  ipoints, lpnt, ipntr,  
     .          a, iwrk(lis+i), iwrk(lic+j), maxn, k, 
     .          shftk, dgrk )
 722        continue 
         do 723 j=1,ncdisc
           slk= slk+c*gkernlsp(points,  ipoints, lpnt, ipntr,  
     .          a, iwrk(lis+i), iwrk(licdisc+j), maxn, k, 
     .          shftk, dgrk )
 723     continue 
         if ( abs(slk) .gt. tol ) then 
            write(6,*) 'equality', i, 'not satisfied', slk
         endif
 720    continue
      do 730 i = 1, nc
         slk=wrk(lr+iwrk(lic+i))
         do 731 j=1,ns
           slk= slk+wrk(ixs+j-1)*gkernlsp(points, ipoints, lpnt, 
     .          ipntr,  a, iwrk(lic+i),  iwrk(lis+j), maxn, k, 
     .          shftk, dgrk )
 731      continue  
         do 732 j=1,nc
           slk= slk+c*gkernlsp(points,  ipoints, lpnt, ipntr, 
     .          a, iwrk(lic+i), iwrk(lic+j), maxn, k, 
     .          shftk, dgrk )
 732      continue 
         do 733 j=1,ncdisc
           slk= slk+c*gkernlsp(points,  ipoints, lpnt, ipntr, 
     .          a, iwrk(lic+i), iwrk(licdisc+j), maxn, k, 
     .          shftk, dgrk )
 733     continue 
         if ( slk .gt. tol ) then 
            write(6,*) 'inequality >', i, 'not satisfied', slk
         endif
 730    continue  
      do 740 i = 1, n0
         slk=wrk(lr+iwrk(li0+i))
         do 741 j=1,ns
           slk= slk+wrk(ixs+j-1)*gkernlsp(points,  ipoints, lpnt, 
     .          ipntr,  a, iwrk(li0+i), iwrk(lis+j), maxn, k, 
     .          shftk, dgrk )
 741      continue  
         do 742 j=1,nc
           slk= slk+c*gkernlsp(points, ipoints, lpnt, ipntr, 
     .           a, iwrk(li0+i), iwrk(lic+j), maxn, k, 
     .          shftk, dgrk )
 742        continue 
         do 743 j=1,ncdisc
           slk= slk+c*gkernlsp(points, ipoints, lpnt, ipntr, 
     .           a, iwrk(li0+i), iwrk(licdisc+j), maxn, k, 
     .          shftk, dgrk )
 743     continue 
         if ( slk .lt. -tol ) then 
            write(6,*) 'inequality <', i, 'not satisfied', slk
         endif
 740    continue 

      do 1750 i = 1, ncdisc
         slk=wrk(lr+iwrk(licdisc+i))
         do 1751 j=1,ns
           slk= slk+wrk(ixs+j-1)*gkernlsp(points,  ipoints, lpnt, 
     .          ipntr,  a, iwrk(licdisc+i), iwrk(lis+j), maxn, k, 
     .          shftk, dgrk )
 1751      continue  
         do 1752 j=1,nc
           slk= slk+c*gkernlsp(points, ipoints, lpnt, ipntr, 
     .           a, iwrk(licdisc+i), iwrk(lic+j), maxn, k, 
     .          shftk, dgrk )
 1752        continue 
         do 1753 j=1,ncdisc
           slk= slk+c*gkernlsp(points, ipoints, lpnt, ipntr, 
     .           a, iwrk(licdisc+i), iwrk(licdisc+j), maxn, k, 
     .          shftk, dgrk )
 1753    continue 
         if ( slk .gt. tol ) then 
            write(6,*) 'disc inequality >', i, 'not satisfied', slk
         endif
 1750    continue
      do 1760 i = 1, n0disc
         slk=wrk(lr+iwrk(li0disc+i))
         do 1761 j=1,ns
           slk= slk+wrk(ixs+j-1)*gkernlsp(points,  ipoints, lpnt, 
     .          ipntr,  a, iwrk(li0disc+i), iwrk(lis+j), maxn, k, 
     .          shftk, dgrk )
 1761      continue  
         do 1762 j=1,nc
           slk= slk+c*gkernlsp(points, ipoints, lpnt, ipntr, 
     .           a, iwrk(li0disc+i), iwrk(lic+j), maxn, k, 
     .          shftk, dgrk )
 1762        continue 
         do 1763 j=1,ncdisc
           slk= slk+c*gkernlsp(points, ipoints, lpnt, ipntr, 
     .           a, iwrk(li0disc+i), iwrk(licdisc+j), maxn, k, 
     .          shftk, dgrk )
 1763    continue 
         if ( slk .lt. -tol ) then 
            write(6,*) 'disc inequality <', i, 'not satisfied', slk
         endif
 1760    continue
        endif

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
ccccccc
ccccccc      end of check
ccccccc
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
       if (n .ge. maxn .or. testerr) goto 999
       itertotal=itertotal+iter
       if (iprint .gt. 0 ) then
         write(6,*)  'maj. iter=', iter1, ' iter=', itertotal, ' y=', y, 
     *               ' ns=', ns, ' n=', n, 
     *               ' nc=', nc,   'n0disc=', n0disc, 
     *               ' ncdisc=', ncdisc
       endif
c
c  compute the minimum slack for the points that have not been included yet.
c
        inminslk=0
        minslk=big
        nremain=maxn-n
c
c  we may try a batch mode when we want to split the data in a few batches,
c  and add the batches incrementally. This does not seem to be more efficient then
c  having all points at once, but we will keep the option in the code for now.
c
        if (batch) then 
          do 131 iter2=1, min(nbatch,nremain)
            i=iwrk(lirm+iter2)
            call addrow( points,  ipoints, lpnt, ipntr, a,  
     *                   wrk(iqs), iwrk(iiqs),  i,
     *                   iwrk(iis), ns, maxn, maxns, k, shftk, dgrk )
            val=0.0d0
            do 132 j = 1, nc
              val=val+gkernlsp( points,  ipoints, lpnt, ipntr, a, i, 
     .                   iwrk(lic+j), maxn, k, shftk, dgrk )
 132        continue
            do 133 j = 1, ncdisc
              val=val+gkernlsp( points, ipoints, lpnt, ipntr, a, i,  
     .                     iwrk(licdisc+j), maxn, k, shftk, dgrk )
 133        continue
            val=val*c
            wrk(lqc+iwrk(lirm+iter2))=val
            n = n+1
            x(n)=0.0d0
            n0=n0+1
            iwrk(li0+n0)=n 
            iwrk(li0ind+n0)=0
 131      continue  
          lirm=lirm+iter2-1
        else
          do 130 iter2=1, nremain
            i=iwrk(lirm+iter2)
            call addrow( points,  ipoints, lpnt, ipntr, a,  
     *                   wrk(iqs), iwrk(iiqs),  i,
     *                   iwrk(iis), ns, maxn, maxns, k, shftk, dgrk )
            call getslk( slk, iwrk(lirm+iter2),  points,   
     *                 ipoints, lpnt, ipntr,a, y, x, c,
     *                 iwrk(iic), iwrk(iicdisc), nc, ncdisc,
     *                 iwrk(iis), ns, iwrk(iiqs),wrk(iqs), maxns, 
     *                 maxn, k, shftk, dgrk, 
     *                 wrk(lqc+iwrk(lirm+iter2)))
            if ( slk .gt. -tol ) then
              n = n+1
              x(n)=0.0d0
              n0=n0+1
              iwrk(li0+n0)=n 
              iwrk(li0ind+n0)=0
            else
              minslk=slk
              inminslk=iter2 
              goto 135
            endif
 130      continue  
        
c
c  if the minimum slack is negative, then add the new point to the active set and
c  resolve the problem
c    

 
 135      if ( minslk .lt. -tol ) then  
            i=iwrk(lirm+inminslk)
            iwrk(lis+ns+1)=i
            ns=ns+1
            n=n+1
            lirm=lirm+iter2

            if ( fullmem .eq. 2 ) then 
              call addcolqs( points,  ipoints, lpnt, ipntr, a,   
     .                    wrk(iqs), iwrk(liqs+ns), iwrk(iis), ns, 
     .                    n, maxn, maxns, k, shftk, dgrk,wrk, .false.) 
            else 
              call addcolqsmin( points,  ipoints, lpnt, ipntr, a, 
     .                       wrk(iqs), iwrk(liqs+ns), 
     .                      iwrk(iis), iwrk(iic), iwrk(ii0), ns, 
     .                      nc, n0,  maxn, maxns, 
     .                      k, shftk, dgrk) 
           endif
           iwrk(iperm+ns-1)=ns
           ibeg=nrnk+1
           initlz=.true.
           wrk(ixs+ns-1)=0.0d0

          else
            nremain=maxn-n
            goto 999
          endif
        endif
       
 500  continue  
cccccccccccccccccccccccccc
c END OF MAIN ITERATION
cccccccccccccccccccccccccc   

 999  continue
c
c
ccccccccccccccccccccccccccccccccccc
c
c  TESTING ERROR COMPUTATION
c
ccccccccccccccccccccccccccccccccccc
       if (testerr) then
        nposmis=0
        nnegmis=0
        nposcor=0
        nnegcor=0
        do 800 i=1, maxn-n
          slk=0  
          do 810 j=1,ns
            slk= slk+wrk(ixs+j-1)*gkernlsp(points, ipoints, lpnt,    
     .          ipntr, a, iwrk(lirm+i), iwrk(lis+j), maxn, k, shftk, 
     .          dgrk )
 810      continue  
         do 820 j=1,nc
           slk= slk+c*gkernlsp(points,  ipoints, lpnt, ipntr,  
     .          a, iwrk(lirm+i), iwrk(lic+j), maxn, k, 
     .          shftk, dgrk )
 820     continue
         slk=slk/a(iwrk(lirm+i))-y 
         if ( slk .gt. 0 ) then
           if (a(iwrk(lirm+i)).lt. 0 ) then 
             nnegmis=nnegmis+1
           else
             nposcor=nposcor+1
           endif
         else
           if (a(iwrk(lirm+i)).lt. 0 ) then 
             nnegcor=nnegcor+1
           else
             nposmis=nposmis+1
           endif
         endif
 800    continue
c       write(6,*) 'nnegmis=', nnegmis, ' nposmis=', nposmis,
c     .            ' nnegcor=', nnegcor, ' nposcor=', nposcor 
c       write(44,*) 'nnegmis=', nnegmis, ' nposmis=', nposmis, 
c     .             ' nnegcor=', nnegcor, ' nposcor=', nposcor 
       endif
cccccccccccccccccccccccccccccccccccccccccccc
c
c  END OF TESTING ERROR COMPUTATION
c
cccccccccccccccccccccccccccccccccccccccccccc        
c
c   count the number of correctly guessed indices during the ipm presolve
c
      nns=0
      nn0=0
      nnc=0
      do 200 i=1,ns
        if (iwrk(itest+iwrk(lis+i)).eq.1) nns=nns+1
 200  continue  
      do 210 i=1,n0
        if (iwrk(itest+iwrk(li0+i)).eq.0) nn0=nn0+1
 210  continue  
      do 220 i=1,n0disc
        if (iwrk(itest+iwrk(li0disc+i)).eq.0) nn0=nn0+1
 220  continue  
      do 230 i=1,nc
        if (iwrk(itest+iwrk(lic+i)).eq.2) nnc=nnc+1
 230  continue 
      do 240 i=1,ncdisc
        if (iwrk(itest+iwrk(licdisc+i)).eq.2) nnc=nnc+1
 240  continue 

c      write(6,*) 'correctly guessed elements of Is=', nns,
c     .           'out of initial', nsapx, 'and out of true',  ns
c      write(6,*) 'correctly guessed elements of I0=', nn0, 
c     .           'out of initial ', n0apx, 'and out of true ', n0+n0disc
c      write(6,*) 'correctly guessed elements of Ic=', nnc, 
c     .           'out of initial ', ncapx, 'and out of true ', nc+ncdisc
c
c      write(6,*) 'number of kernel calls=', ncalls
c      ncalls=0
      if (iprint. ge. 0 ) then 
        write(6,*) 'Number of active sup. vec.', ns
        write(6,*) 'Number of bounded sup.vec.', nc+ncdisc
        write(6,*) 'Bias=', -y  
        if (iprint.ge. 3) then 
          write(6,*) 'List of multipliers that are between 0 and C:'
          do 300 i=1, maxn
            if (x(i) .gt. tol .and. x(i) .lt. c-tol) then
              write(6,*) i, ' ', x(i)
            endif
 300      continue
          write(6,*) 'List of indices of multipliers that at C:'
          do 310 i=1, maxn
            if (x(i) .ge. c-tol) then
              write(6,*) i
            endif
 310      continue
         endif
            
Ctime      totalt=mclock()-currt
        timek=0.0d0
Ctime        write(6,*) 'cpu time=', totalt/100
c      write(44,*) 'cpu time=', totalt/100
Ctime        totalt=mclock()
c       write(44,*) 'correctly guessed elements of Is=', nns,
c     .           'out of initial', nsapx, 'and out of true',  ns
c       write(44,*) 'correctly guessed elements of I0=', nn0, 
c     .           'out of initial ', n0apx, 'and out of true ', n0+n0disc
c       write(44,*) 'correctly guessed elements of Ic=', nnc, 
c     .           'out of initial ', ncapx, 'and out of true ', nc+ncdisc
c
c       write(44,*) 'number of kernel calls=', ncalls
c       write(44,*) 'ns=', ns, 'nc=', nc+ ncdisc
c       write(44,*) 'y=', y 
        

        do 9935 i=1, ns   
          call addcolqs(points,  ipoints, lpnt, ipntr, a, 
     .                    wrk(iqs), iwrk(liqs+i), iwrk(iis), 
     .                    ns, n, maxn, maxns,  k, shftk, dgrk,
     .                     wrk, .false.)
 9935   continue  
Ctime       totalt=mclock()-totalt
Ctime       write(6,*) 'cpu time to check optimality=', totalt/100
c      write(6,*) 'kernel time=', timek/100
c      write(44,*) 'cpu time to check optimality=', totalt/100
c      write(44,*) 'kernel time=', timek/100
      endif
      return
 1000 format( 'Insufficient real memory. Need at least', 1x, I10 )
 1100 format( 'Insufficient integer memory. Need at least', 1x, I10 )
 4000 format( 'bias=', D14.7)
       end


      subroutine discpts(iset, nset, indic, idisc, ndisc,  x, val,
     *                   maxn, thresh)
      integer          iset(nset), nset, indic(nset), idisc(maxn), 
     *                 ndisc, thresh, maxn
      double precision x(maxn), val
c
c  local variables
c
      integer i, l, j
c
c  if an indicator exceeds the threshold then we skip the appropriate variable
c
        l=1
        do 10 i=1, nset
          if ( indic(l) .gt. thresh ) then
            x(iset(l))=val
            idisc(ndisc+1)=iset(l)
            ndisc=ndisc+1
            do 11 j=l, nset-1-(i-l)
              iset(j)=iset(j+1)
              indic(j)=indic(j+1)
 11        continue
          else
            l=l+1
          endif
 10    continue
       nset=l-1
       return
       end  


 

       subroutine getslk( slk, i, points,  ipoints, lpnt, ipntr,  
     *                   a, y, x, c, ic, icdisc, nc,  
     *                   ncdisc, is, ns, iqs, qs, maxns,  maxn, k,
     *                   shftk, dgrk, val)
       integer          i, ic(nc),nc , icdisc(ncdisc), ncdisc,maxn,k,
     *                  dgrk, is(ns), iqs(ns),  maxns, ns, idisc, lpnt,
     *                  ipoints(lpnt), ipntr(maxn)
       double precision slk, points(lpnt), a(maxn), y, c, x(maxn), 
     *                  shftk, qs(maxn, maxns), val
      double precision  gkernlsp
      external          gkernlsp
c
c  local variables
c
       integer          j, l
       slk=0.0d0
       do 10 j = 1, nc
         slk=slk+gkernlsp( points,  ipoints, lpnt, ipntr, a, i, 
     .                  ic(j), maxn, k, shftk, dgrk )
 10    continue
       do 20 j = 1, ncdisc
         slk=slk+gkernlsp( points, ipoints, lpnt, ipntr, a, i,  
     .                  icdisc(j), maxn, k, shftk, dgrk )
 20    continue
       slk=slk*c
       val=slk
       slk=slk-1.0d0 -a(i)*y 
       do 30 j=1,ns
         slk= slk+x(is(j))*qs(i, iqs(j))
 30    continue
       return
       end


