cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c  SVM-QP, written by Katya Scheinberg, IBM Research, 2005.
c  to contact the author email at katyas'at'us'dot'ibm'dot'com
c
c Copyright (C) 2005 International Business Machines.
c All Rights Reserved.
c This code is published under the Common Public License.c

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine qp(lowrnk, points, ldp,  perm, a, c, r, b, x, s, y, xi,  
     *              p, beta, lambda, n, k, dx, ds, dy, dxi,  wrk, lwrk)
c
c  subroutine parameters
c
      integer           n, k, lwrk, perm(n), ldp, info
      double precision  a(n), x(n), points(ldp,k), c, b, r(n),   
     *                  p(n, k)   , beta(n, k), lambda(n), s(n), y, 
     *                  xi(n), dx(n), ds(n), dy, dxi(n), wrk(lwrk)
      logical           lowrnk
c
c  a      - n-dimensional array of 1's and -1's
c  points - n by k array of data points
c  c      - feasibility parameter, 
c  r      - linear part of the objective function
c  x      - the primal solution (which we used to call dual solutions alpha)
c  s      - dual slack
c  y      - dual variable = negative bias
c  xi     - dual variable resposible for infeasibility (xi(i)=0 if 
c           the i-th point is separated correctly)
c  p      - n by k array of k vectors p^i that contain factorization
c           information
c  beta   - n by k array of k vectors beta^i that contain factorization
c           information
c  lambda - n-dimensional array - diagonal elements of the diagonal
c           matrix of the Cholesky factorization
c  dx, dy, ds, dxi - vectors on each iteration containg step for
c           the appropriate vector of variables (x, y, s, xi).
c           These arrays are not needed neither as input nor as output,
c           they are only used in qp() internally, I use them as parameters,
c           just for convinience, to be able to write them as separate vectors,
c           and not as parts of wrk array.
c  wrk    - working space
c  lwkr   - lenght of working space
c
c
c  local variables
c 
      double precision  alpha, alpha1, alpha2, one, zero, rb,
     *                  rhs, schcomp, primal_infeas, dual_infeas, gap,
     *                  mu, newmu, sigma, tol, tmp, mxslk, totalt,
     *                  timediff, primal_obj, dual_obj
      integer           iwr, iwrc, iwr1, iwppx, iwpx, iwppdx, icurw,
     *                  lwr, lwrc, lwr1, lwppx, lwpx, lwppdx, maxit, 
     *                  iter, i      
      parameter        (one=1.0d0, zero=0.0d0)
      double precision  ddot, dnorm
      external          ddot, dnorm
      intrinsic         min, abs
      totalt=0.
 
c
c  memory  allocation for additional auxilliary arrays
c
      iwr    = 1
      iwrc   = iwr+n
      iwr1   = iwrc+n
      iwppx  = iwr1+n
      iwpx   = iwppx+n
      iwppdx = iwpx+k
      icurw  = iwppdx+n

      lwr    = iwr - 1
      lwrc   = iwrc - 1
      lwr1   = iwr1 - 1
      lwppx  = iwppx - 1
      lwpx   = iwpx - 1
      lwppdx = iwppdx - 1
c
c  Initialization
c
c
c  maximum number of iterations
c
      maxit = 100
c
c  absolute tolerance (stopping criteria)
c
      tol   = 1.0d-10
      if (lowrnk) then
c 
c  permuting the data
c  
       do 3 i=1,n
         call dcopy(k, points(i, 1 ), ldp,  p(perm(i), 1), n)
 3     continue  
       do 4 i=1,n
        call dcopy(k,  p(i, 1 ), n, points(i, 1), ldp)
 4     continue

c  
c
c  Check if current point is optimal
c
       do 1030 i=1,k    
         wrk(lwpx+i)=ddot(n, points(1, i), 1, x, 1)
 1030  continue
c
c  Now compute PPx=Points*Px
c
       do 1040 i=1,n   
        wrk(lwppx+i)=ddot(k, points(i, 1), ldp, wrk(iwpx), 1)
 1040  continue
      else
        do 1031 i=1,k    
          wrk(lwppx+i)=ddot(n, points(1, i), 1, x, 1)
 1031   continue
      endif
c
c  compute vector e-PPx+a*y and find its largest negative component
c  Recall that -PPx+a*y+s-xi=-e, by setting s>max(-e+PPx-a*y) we 
c  guarantee that xi>0 and thus we have an intial feasible solution.
c
      mxslk=zero
      do 1050 i=1,n
        wrk(lwr1+i)=-r(i)-wrk(lwppx+i)+a(i)*y
        if ( mxslk .lt. -wrk(lwr1+i)) mxslk = -wrk(lwr1+i)
 1050   continue
      mxslk=mxslk+0.1d0
      primal_obj=-0.5d0*ddot(n, x, 1, wrk(iwppx), 1) 
      dual_obj=-primal_obj
c
c  set s and xi accordingly the "dual risidual" -e+PPx-a*y-s+xi=0
c
      rb = b - ddot(n, a, 1, x, 1)       
c
c  initialization
c     set x(i)=0.1 for all i
c         y   =0.0
c 
      do 20 i=1, n
        x(i)=0.5d0
 20   continue
      y=zero  
c
c  This computes Points*Points\t*x. First compute Px=Points\t*x
c
      if (lowrnk) then
        do 30 i=1,k    
          wrk(lwpx+i)=ddot(n, points(1, i), 1, x, 1)
 30     continue
c
c  Now compute PPx=Points*Px
c
        do 40 i=1,n   
          wrk(lwppx+i)=ddot(k, points(i, 1), ldp, wrk(iwpx), 1)
 40     continue
      else
        do 31 i=1,k    
          wrk(lwppx+i)=ddot(n, points(1, i), 1, x, 1)
 31     continue
      endif
c
c  compute vector e-PPx+a*y and find its largest negative component
c  Recall that -PPx+a*y+s-xi=-e, by setting s>max(-e+PPx-a*y) we 
c  guarantee that xi>0 and thus we have an intial feasible solution.
c
      mxslk=zero
      do 50 i=1,n
        wrk(lwr1+i)=-r(i)-wrk(lwppx+i)+a(i)*y
        if ( mxslk .lt. -wrk(lwr1+i)) mxslk = -wrk(lwr1+i)
 50   continue
      mxslk=mxslk+0.1d0
c
c  set s and xi accordingly the the "dual risidual" -e+PPx-a*y-s+xi=0
c
      do 55 i=1,n
        s(i)=mxslk
        xi(i)=wrk(lwr1+i)+s(i)
        wrk(lwrc+i)=zero
 55   continue
      rb = b - ddot(n, a, 1, x, 1) 

c
c  compute the initial parameter mu as (x'*s+xi'*(c-x))/2n
c
      mu=zero
      do 57 i=1,n
        mu=mu+x(i)*s(i)+xi(i)*(c-x(i))
 57   continue
      mu=mu/(2*n) 


c
c  Begin main iteration
c
      do 1000 iter=1,maxit
c
c  Compute the diagonal matrix D=SX^(-1)+Xi(C-X)^(1)
c

      do 60 i=1,n
        wrk(icurw+i-1)=s(i)/x(i)+xi(i)/(c-x(i))
 60   continue
      if (lowrnk) then    
c
c  compute the product form factorization of D+Points*Points\t
c  
        call pfcf(p, beta, lambda, wrk(icurw), points, ldp, n, k)
      else
        do 64 i=1,n
          call dcopy(k, points(i, 1 ), ldp,  p(i, 1), n)
          p(i,i)=p(i,i)+wrk(icurw+i-1)
 64     continue
        call  dpotrf('U', n, p, n, info)
      endif
 
c
c  compute predictor step from
c     -Q*dx + a*dy + ds - dxi = rc
c      a'*dx                  = rb
c      X*ds + S*dx            = -Xs
c      -Xi*dx+(C-X)*dxi       = -Xi(C-X)
c
c  From above r1=e-a*y+Q*x
c     compute r1=Q^(-1)*r1 (by using the factorization of Q computed
c                           by pfcf) 
      if ( lowrnk ) then    
        call slvsys(p, beta, lambda, wrk(iwr1), n, k)
      else
        call dpotrs('U', n, 1, p, n, wrk(iwr1), n, info)
      endif
c
c  compute r=Q^(-1)*a
c
      call dcopy(n, a, 1, wrk(iwr), 1)
      if ( lowrnk ) then 
        call slvsys(p, beta, lambda, wrk(iwr), n, k)
      else
        call dpotrs('U', n, 1, p, n, wrk(iwr), n, info)
      endif

c
c  compute a\t*Q^(-1)*a
c
      schcomp=ddot(n, a, 1, wrk(iwr), 1)
c
c   rhs=rb-a'*r1
c 
      rhs=rb-ddot(n, a, 1, wrk(iwr1), 1)
c
c  dy=rhs/(a\t*Q^(-1)*a)
c
      dy=rhs/schcomp
c
c   scale r=Q^(-1) by dy: new  r=dy*Q^(-1)*a
c
      call dscal(n, dy, wrk(iwr), 1)        
c
c  dx=r+r1
c
      do 80 i=1, n
        dx(i)=wrk(lwr+i)+wrk(lwr1+i)
 80   continue
c
c  Px=Points\t*dx
c  
      if ( lowrnk ) then
        do 90 i=1,k    
          wrk(lwpx+i)=ddot(n, points(1, i), 1, dx, 1)
 90   continue
c
c  PPdx=Points*Px
c        
        do 100 i=1,n  
          wrk(lwppdx+i)=ddot(k, points(i, 1), ldp, wrk(iwpx), 1)
 100    continue
      else
        do 105 i=1,n  
          wrk(lwppdx+i)=ddot(k, points(i, 1), ldp, dx, 1)
 105    continue
      endif
c
c  ds  = -S*X^(1)*dx-s
c  dxi = -rc+a*dy-PPdx+ds
c
      do 110 i=1,n
        ds(i)=-s(i)/x(i)*dx(i)-s(i)
        dxi(i)=-wrk(lwrc+i)+a(i)*dy-wrk(lwppdx+i)+ds(i)
 110  continue
c
c  Now that we have computed dx, dy, ds, dxi we need to compute the 
c  step length alpha<=1. The step length should be such that x+alpha*dx>=0
c  x+alpha*dx<=c, s+alpha*ds>=0 and xi+alpha*dxi>=0
c
c compute the step lenght
c
      alpha1=one
      alpha2=one
      do 120 i=1,n
        if ( dx(i).lt. 0 .and. alpha1 .gt. -x(i)/dx(i) ) then
          alpha1=-x(i)/dx(i)
        endif
        if ( dx(i).gt. 0 .and. alpha1 .gt. (c-x(i))/dx(i) ) then
          alpha1=(c-x(i))/dx(i)
        endif
        if ( ds(i).lt. 0 .and. alpha2 .gt. -s(i)/ds(i) ) then
          alpha2=-s(i)/ds(i)
        endif
        if ( dxi(i).lt. 0 .and. alpha2 .gt. -xi(i)/dxi(i) ) then
          alpha2=-xi(i)/dxi(i)
        endif
 120  continue
      alpha=min(alpha1,alpha2)  
c
c  Estimate the new value for mu
c
      newmu=zero
      do 130 i=1,n
        newmu=newmu+(x(i)+alpha*dx(i))*(s(i)+alpha*ds(i))+
     *        (xi(i)+alpha*dxi(i))*(c-x(i)-alpha*dx(i))
 130  continue
      newmu=newmu/(2*n)
c
c  parameter sigma (to be used in the corrector step)
c  
      sigma=(newmu/mu)**3

c
c  compute corrector step
c 
c  Let del(x)=dx, del(s)=ds and del(xi)=dxi from the predictor step
c  we compute new dx, ds, dxi, dy from
c
c     -Q*dx + a*dy + ds - dxi = rc
c      a\t*dx                  = rb
c      X*ds + S*dx            = sigma*mu*e-Del(s)*del(x)-X*s
c      -Xi*dx+(C-X)*dxi       = sigma*mu*e+Del(xi)*del(x)-Xi(C-X)
c
c
c  Compute r1=X^(-1)*(-sigma*mu*e+Del(s)*del(x)+X*s) +
c             (C-X)^(1)*(sigma*mu*e+Del(xi)*del(x)-Xi(C-X))-e-PPx-a*y-s+xi
c
      do 140 i=1,n
        wrk(lwr1+i)=(one/x(i))*(-sigma*mu+dx(i)*ds(i)+x(i)*s(i))
     *      +(one/(c-x(i)))*(sigma*mu+dx(i)*dxi(i)-(c-x(i))*xi(i))
     *      +r(i)+wrk(lwppx+i)-a(i)*y-s(i)+xi(i)
 140  continue
c
c  r1=Q^(-1)*r1
c
      if ( lowrnk ) then    
        call slvsys(p, beta, lambda, wrk(iwr1), n, k)
      else
        call dpotrs('U', n, 1, p, n, wrk(iwr1), n, info)
      endif
c
c  rhs = a\t*r1
c
      rhs=rb+ddot(n, a, 1, wrk(iwr1), 1)
c
c  scale r by 1/dy, so that r=Q^(-1)*a
c
      call dscal(n, one/dy, wrk(iwr), 1)    
c
c  dy=rhs/(a'*Q^(-1)*a)
c
      dy=rhs/schcomp
c
c  scale r by the new dy, so r=Q^(1)*a*dy
c 
      call dscal(n, dy, wrk(iwr), 1)        
c
c  dx=r-r1
c  ds=X^(-1)*(sigma*mu*e-Del(s)del(x)-S*dx)-s
c
      do 150 i=1, n
        tmp=wrk(lwr+i)-wrk(lwr1+i)
        ds(i)=(sigma*mu-dx(i)*ds(i)-s(i)*tmp)/x(i)-s(i)
        dx(i)=tmp
 150   continue
c
c  Px=Points\t*dx
c 
      if ( lowrnk ) then 
        do 160 i=1,k    
          wrk(lwpx+i)=ddot(n, points(1, i), 1, dx, 1)
 160    continue
c
c  PPdx=Points*Px
c
        do 170 i=1,n   
          wrk(lwppdx+i)=ddot(k, points(i, 1), ldp, wrk(iwpx), 1)
 170    continue
      else
        do 175 i=1,n   
          wrk(lwppdx+i)=ddot(k, points(i, 1), ldp, dx, 1)
 175   continue
      endif
c
c  dxi=-rc+a*dy-PPdx-ds
c
      do 180 i=1,n
        dxi(i)=-wrk(lwrc+i)+a(i)*dy-wrk(lwppdx+i)+ds(i)
 180  continue  

c
c  compute the step lenght alpha
c  The step lenght is maximum possible, such that all variables (except y)
c  stay within bounds (see above, for predictor step)
c
      alpha1=1000*one
      alpha2=1000*one
      do 190 i=1,n
        if ( dx(i).lt. 0 .and. alpha1 .gt. -x(i)/dx(i) ) then
          alpha1=-x(i)/dx(i)
        endif
        if ( dx(i).gt. 0 .and. alpha1 .gt. (c-x(i))/dx(i) ) then
          alpha1=(c-x(i))/dx(i)
        endif
        if ( ds(i).lt. 0 .and. alpha2 .gt. -s(i)/ds(i) ) then
          alpha2=-s(i)/ds(i)
        endif
        if ( dxi(i).lt. 0 .and. alpha2 .gt. -xi(i)/dxi(i) ) then
          alpha2=-xi(i)/dxi(i)
        endif
 190  continue
c
c  we take 99% of the step to the boundary
c
      alpha=min(one, 0.99d0*min(alpha1,alpha2))
c
c  update x=x+alpha*dx, s=s+alpha*ds, xi=xi+alpha*dxi
c
      call  daxpy(n, alpha, dx, 1, x, 1)
      call  daxpy(n, alpha, ds, 1, s, 1)
      call  daxpy(n, alpha, dxi, 1, xi, 1)
      y=y+alpha*dy
c
c  compute new PPx=Points*Points\t*x; new x= old x + alpha * dx, thus
c  new PPx= old PPx + alpha PPdx
c
       do 200 i=1,n   
        wrk(lwppx+i)=wrk(lwppx+i)+ alpha*wrk(lwppdx+i)
 200   continue
c
c  compute new r1, rc and rb
c
      do 220 i=1,n
        wrk(lwr1+i)=-r(i)-wrk(lwppx+i)+a(i)*y
        wrk(lwrc+i)=-wrk(lwr1+i)-s(i)+xi(i)
 220  continue
      rb = b - ddot(n, a, 1, x, 1)
c
c  primal objective= 0.5*x*PPx-e*x, dual objective=0.5*x*PPx+c*\sum(xi), 
c
      primal_obj=-0.5d0*ddot(n, x, 1, wrk(iwppx), 1) 
      dual_obj=-primal_obj-b*y
      do 222 i=1, n
        primal_obj= primal_obj - r(i)*x(i)
        dual_obj  = dual_obj + c*xi(i) 
 222  continue  
c
c  primal and dual infeasibilities
c
      dual_infeas=dnorm(n, wrk(iwrc), 1)
      primal_infeas=abs(rb)
c
c  duality gap=x\t*s+xi\t*(c-x)
c
      gap=zero
      do 230 i=1,n
        gap=gap+x(i)*s(i)+xi(i)*(c-x(i))
 230  continue
c
c  new value for parameter mu: gap/2n
c
      mu=gap/(2*n)
c
c  check for convergence, if converged, exit
c
      if (gap/(abs(primal_obj)+1+dnorm(n,r,1))+dual_infeas+
     .    primal_infeas .lt. tol) 
     .  then
        goto 2000
      endif
c
c  print iteration info
c
c      write(6,3000) iter, alpha, gap, dual_infeas, primal_infeas, 
c     *              totalt/100, primal_obj, dual_obj
c      write(44,3000) iter, alpha, gap, dual_infeas, primal_infeas,
c     *               totalt/100, primal_obj, dual_obj
1000  continue 
 
2000   continue
c      write(6,*) 'finished'
c      write(6,*) 'gap=', gap
c      write(6,*) 'bias=', y
c      write(6,*) 'primal infeas=', primal_infeas
c      write(6,*) 'dual infeas=', dual_infeas
c      write(6,*) 'primal objective=', primal_obj
c      write(6,*) 'dual objective=', dual_obj
c      write(44,*) 'finished'
c      write(44,*) 'gap=', gap
c      write(44,*) 'bias=', y
c      write(44,*) 'primal infeas=', primal_infeas
c      write(44,*) 'dual infeas=', dual_infeas
c      write(44,*) 'primal objective=', primal_obj
c      write(44,*) 'dual objective=', dual_obj
c      timediff=mclock()-totalt
c      totalt=totalt+timediff      
c      write(6,*) 'cpu time=', totalt/100
c      write(44,*) 'cpu time=', totalt/100
      return
3000  format(I4, 1x, d12.5, 1x, d12.5, 1x, d12.5, 1x, d12.5, 1x, d12.5,
     *       1x, d12.5, 1x, d12.5)
      end
c  **********************************************************************
c
c  Subroutine PFCF for computing the product form Cholesky factorization
c
c  **********************************************************************
      subroutine pfcf(p, beta, lambda, d, points, ldp, n, k)
c
c   This subroutine computes factorization of the form
c     L_1*L_2*....L_k*Lambda*L_k'*...*L_2'*L_1'
c   of the matrix
c     D+Points*Points', where D is diagonal n by n and Points is n by k
c       

c
c  parameters
c
      double precision p(n,k), beta(n,k), lambda(n), d(n), points(ldp,k)
      integer          n, k, ldp
c
c  local variables
c
      integer          i, j
      double precision t, tbar, one
      parameter       (one = 1.0d0)
c
c  set Lambda=D
c
      call dcopy(n, d, 1, lambda, 1)
c
c  for i from 1 to k
c
      do 100 i=1,k
c
c   vector p^i=Points(1,i) - i-th column of Points
c
        call dcopy(n, points(1, i), 1, p(1, i), 1)
c
c   solve p^i=L_j^(-1)*....L_2^(-1)*L_1^(-1)*p^i
c   (the necessary infomation was already computed
c   for L_j, j<i)
c
c
        do 10 j=1,i-1
          call solvelt(p(1,j), beta(1,j), p(1,i), n, 1, 1)
 10     continue
c
c  apply recursion formulas to find vector beta^i and new lambda
c
        t=one
        do 20 j=1,n
          tbar=t+p(j, i)**2/lambda(j)
          beta(j, i)=p(j,i)/(lambda(j)*tbar)
          lambda(j)=lambda(j)*tbar/t
          t=tbar
 20     continue  
 100  continue

      return
      end
c  **********************************************************************
c
c  Subroutine SLVSYS for solving system with known product form 
c  factorization
c
c  **********************************************************************  

      subroutine slvsys(p, beta, lambda, r, n, k)
c
c  Solving for r=M^(-1)*r, where
c    M=L_1*L_2*....L_k*Lambda*L_k\t*...*L_2\t*L_1\t
c
c
c  parameters
c
      double precision p(n,k), beta(n,k), lambda(n), r(n)
      integer          n, k
c
c  local variables
c
      integer          i, j

c
c  Solve  system r=L_i^(-1)*r
c
      do 10 i=1, k
        call solvelt(p(1,i), beta(1,i), r, n, 1, 1)
 10   continue
c
c  r=Lambda^(-1)*r
c
      do 20 i=1,n
        r(i)=r(i)/lambda(i) 
 20   continue
c
c  Solve  system r=(L_i^(-1))\t*r (for the transpose of L_i)
c     
      do 30 i=k, 1, -1
        call solveltt(p(1,i), beta(1,i), r, n, 1, 1)
 30   continue

      return
      end

      subroutine solvelt(p, beta, r, nrow, ibp,ibq)
c
c  solve for r=L^(-1)*r  where L_ij=p_i*beta_j for i<j,  L_ii=1, L_ij=0 i>j
c
c  parameters
c
       double precision   beta(nrow), p(nrow), r(nrow)
       integer            ibp, ibq, nrow
c
c local variables
c
       double precision sum
       integer          ibeg, i
c
c  Here ibp and ibq are always equal to 1.
c

       if (ibp.lt.ibq) then
         ibeg=ibq
       else
         ibeg=ibp
       endif
       sum=r(ibeg)*beta(ibeg)
       
       do 20 i=ibeg+1,nrow        
         r(i)=r(i)-p(i)*sum
         sum=sum+r(i)*beta(i)
20     continue
       return
       end


       subroutine solveltt(p, beta, r, nrow, ibp,ibq)
c 
c   solve for transpose system r=L\^(-1)*r  where L_ij=p_i*beta_j for i<j, 
c                                                 L_ii=1, Lij=0 for i>j
c
c  parameters
c
       double precision   beta(nrow), p(nrow), r(nrow)
       integer            ibp, ibq, nrow
c
c local variables
c
       double precision sum
       integer          ibeg, i
c
c  ibq and ibp are equal to 1
c
       if (ibp.lt.ibq) then
         ibeg=ibq
       else
         ibeg=ibp
       endif

       sum=r(nrow)*p(nrow)
       do 20 i=(nrow-1),ibeg,-1        
         r(i)=r(i)-beta(i)*sum
         sum=sum+r(i)*p(i)
20     continue
       return
       end

       double precision function dnorm(n, r, ir)
       double precision r(n)
c
c  euclidean norm of vector r
c
       integer          n, ir, i
       intrinsic        sqrt
       dnorm=0.0d0
       do 10 i=1,n
         dnorm=dnorm+r(1+i*ir-1)**2
 10    continue
       dnorm=sqrt(dnorm)
       return
       end  
