cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c  SVM-QP, written by Katya Scheinberg, IBM Research, 2005.
c  to contact the author email at katyas'at'us'dot'ibm'dot'com
c
c Copyright (C) 2005 International Business Machines.
c All Rights Reserved.
c This code is published under the Common Public License.
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine svmqp(x, y, a, points, ipoints, lpnt, 
     .              lipnts, ipntr, lipntr,  c, r, maxn, maxns, 
     .              k, nrnka, dgr, shft, iprint, fullmem, 
     +              ninit, presolve, tol, wrk, lwrk,  iwrk, liwrk)
      integer n, k, i, j, lwrk,  lpnt, lipnts, lipntr, 
     .        ipoints(lipnts), ipntr(lipntr), liwrk,  maxsift0, 
     .        maxsiftc, maxit, maxns, ninit, wpres, iwpres,
     .        iwrk(liwrk), nrnka, dgr, iprint, fullmem, ncalls
      double precision a(maxn),  y, points(lpnt), r(maxn), c, wrk(lwrk),
     .                 shft, tol, x(maxn)
      logical          usesift, sparse, cornrsln, presolve
      double precision timek,  currt, totalt
      common /sprs/    sparse, ncalls, timek
c
c  Solving a convex quadratic optimization problem:
c             min  0.5*x'*Q*x - r'*x
c             s.t. a'*x=0
c                  0 =< x <= C
c                  
c
c   'x' is a n-dimensional vector, 'r' - is an n-vector of all 1's,
c   'a' is an n-vector of  1 and -1 labels.  'Q' - is an n by n matrix,
c   'Q' is the kernel matrix scaled on each side by the diagonal matrix, 
c    whose entries are lablels, 'a'. C is the penalty term.
c    
c
c  The dual problem to the above is:
c     
c     min 0.5*x'*Q*x + C'*x
c     s.t.  -Q*x+y*a+s-xi=-e
c            s>=0, xi>=0
c
c  'xi' is a n-vector, 'y' is a scalar and is equal to the negative bias,
c  i.e. y=-Bias. 
cccccccccccccccccccccccccccccccccccccccccccccccc
c   parameters of the routine
cccccccccccccccccccccccccccccccccccccccccccccccc
c
c   x      - the array of the Lagrange multimpliers (alphas, in the SVM literature)
c   y      - the negative bias of the linear separator (b or beta in SVM literature)
c   a      - the array of labels
c   points - the array containing the data in the sparse format, if the data is sparse
c            and if data is dense, then in dense format, storing one point after the other.
c   ipoints- the idices of the nonzero elements of the data points, if "points" is
c            a sparse array, otherwise ignore
c   lpnt   - the lenght of points
c   lipnts - length of ipoints
c   ipntr  - the array of pointers to the begining of each  data point in "points"
c            and "ipoints". 
c   linpnr - length of array 'ipntr'
c   c      - the penalty parameter
c   r      - the right hand side vector of length 'maxn', whose elements should all equal -1.
c            Subsequent versions will allow for different values of 'r'.
c   maxn   - the total number of points
c   maxns  - the maximum number of active support vectors allowed 
c   k      - the dimension of the data
c   nrnka  - maximum rank of the approximate Hessian used in presolve. See "presolve" details.
c   dgr    - the kernel parameter indicating degree for polynomial kernels (dgr=1 - linear kernel)
c            of indicating the RBF kernel (dgr=-1). Any other kernel has to be coded into
c            kernl.f file and idicated by another value of "dgr". 
c  shft    - the kernel parameter, shift in poly kernels and sigma in RBF kernels.
c  iprint  - print level. >=2, then each iteration info is printed, =1, then each 10th
c            iteration infor is printed, =0, then no iteration info is printed, only the
c            funal output.
c  fullmem - the memory use level. Right now set fullmem=2 at all times.
c  ninit   - number of initial points, if an incremental mode is used.
c  presolve- a logical variable idicating if a presolve of the approximate low-rank
c            problems with an IPM is desired.
c  tol     - tolerance for solution
c  wrk     - double precision working space
c  lwrk    - the length of double precision working space, at least 11*maxn+maxns*maxns+maxn*maxns
c  iwrk    - integer working space
c  liwrk   - the length of integer working space, at least 13*maxn+2*maxns
c  
c
c
cccccccccccccccccccccccccccccccccccccccccc
c  set some other controlling parameters
cccccccccccccccccccccccccccccccccccccccccc
c
c  usesift - whether to use "sifting" technique for choosing the constraints to price
c  maxsift0 - number of constraints with nonzero slack s to be considered at a time
c  maxsift0 - number of constraints with nozero surplus to be considered at a time
c
      usesift = .true.
      maxsift0=50
      maxsiftc=50
c
c  fullmem = 1 - at each iteration store the values of matrix Q that are in the columns corresponding to
c  set Is and with rows in Is, Ic and I0.
c  fullmem = 2 - at each iteration store  the whole columns of Q corresponding to set Is.
c
      if (fullmem .gt. 2 .or. fullmem .lt. 0 ) then
        if (iprint.ge.0) write(6,*) 
     .     'Wrong value for memory option', fullmem
        stop
      endif
      do 5 i=1,maxn
        if ( (a(i).ne. 1) .and. (a(i).ne. -1)) then 
          if(iprint .ge.0) write(6,*) 'ERROR!! 
     .      A label of example', i, 'is',a(i)
          goto 999
        endif
 5    continue 
c
c  cornsln=true - project the approximate solution obtained by IPM presolve tp a corener solution
c
      cornrsln=.false.
c
c  presolve=true - perform a presolve by solving an approximate problem by a product from IPM.
c
      presolve=.true.
      if (nrnka.eq.0) presolve=.false.
      maxit=1000000
c
c  ninit - number of initial point, for incremental mode. If ninit=maxn, then all points are used
c  from the beginning. Sifting is not efficient in the incremental mode.
c
      if ( ninit .eq. 0 )  ninit = maxn
      if ( presolve )  then
        wpres = 12*maxn+2*nrnka+3*maxns*nrnka
        iwpres= maxn
      else
        wpres = 0
        iwpres= 0
      endif
      if ( lwrk .lt. 11*maxn+maxns**2+maxn*maxns+wpres) then 
        if (iprint .ge. 0) write(6, 1000) 
     .    11*maxn+maxns**2+maxn*maxns+wpres
          goto 999
      endif
      if ( liwrk .lt. 13*maxn+2*maxns+iwpres) then 
        if (iprint .ge. 0) write(6, 1010) 
     .    13*maxn+2*maxns+iwpres
          goto 999
      endif
      n=maxn
      icurw=1
      icuriw=1
ccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c    presolve with an approximate problem, if desired c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      if (presolve) then
        ig      = 1 
        itrc    = ig + n*nrnka
        is      = itrc + nrnka
        ixi     = is + n
        id      = ixi + n
        ip      = id + n
        ibeta   = ip +n*nrnka
        ilambda = ibeta + n*nrnka
        idx     = ilambda + n
        ids     = idx + n
        idxi    = ids + n
        ic      = idxi + n
        icurw   = ic + n

        iperm   = 1
        icuriw   = iperm + n
c
c  check if memory is sufficient, if not, print a message and exit
c

Ctime         currt=mclock()

c
c  Find incomplete Cholesky factors
c
        call cholinpiv(a, points, ipoints, lpnt, 
     .                 ipntr,  wrk(ig), wrk(itrc),
     *                 iwrk(iperm), n , k, nrnka, shft, dgr, tol)
c
c adjust the data according to the permutations
c
        do 10 i=1, n
          wrk(icurw+i)=a(iwrk(iperm+i-1))
 10     continue
        do 11 i=1,n
          a(i)=wrk(icurw+i)
          wrk(ic+i-1)=c
 11     continue 

c
c  call the optimizer
c
        call qpipm(wrk(ig), a, wrk(ic), r, x,  
     *          wrk(is), y, wrk(ixi), wrk(ip), wrk(ibeta), wrk(ilambda),
     *          n, nrnka, wrk(idx), wrk(ids), dy, wrk(idxi), 
     *          wrk(icurw), lwrk-icurw+1)




c
c reorder vectors a and x back according to permulation
c
        do 20 i=1, n
          wrk(icurw+iwrk(iperm+i-1)-1)=a(i)
 20     continue
        do 21 i=1,n
          a(i)=wrk(icurw+i-1)
 21     continue  
        do 30 i=1, n
          wrk(icurw+iwrk(iperm+i-1)-1)=x(i)
 30     continue
        do 31 i=1,n
          x(i)=wrk(icurw+i-1)
 31     continue

c
c  release (virtually) the extra memory used in qpipm 
c


        icurw  = 1
        icuriw = 1
ccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c    end of presolve                                  c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccc  
Ctime        totalt=mclock()-currt
Ctime       if (iprint .ge. 0) write(6,*) 'cpu time for presolve=', 
Ctime     .     totalt/100
            if (iprint .ge. 0) write(6,*) 'end of presolve'
c       write(44,*) 'cpu timefor presolve=', totalt/100
      else 
c
c  if no presolve set the initial solution to all zeros
c
        y=1000
        do 15 i=1, maxn
          x(i)=0.0d0
 15     continue  
      endif
ccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c  call the active set solver and solve the problem   c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      call inc_actvset(points, ipoints, lpnt, lipnts, ipntr, 
     .                a, c, x, y,  k, 
     .                nrnk, shft, dgr, maxns, maxn, ninit, usesift, 
     .                maxsift0, maxsiftc, fullmem, tol, maxit, cornrsln, 
     .                iprint, wrk(icurw), lwrk, iwrk(icuriw), liwrk)


c      write(6,*) 'number of kernel calls=', ncalls
 999  continue
      return
 1000 format( 'Insufficient real memory. Need at least', 1x, I10 )
 1010 format( 'Insufficient integer memory. Need at least', 1x, I10 )
 6000 format( 3(I4,/), D8.4, /, D8.4, /,3(I4,/) ) 
      end
